use crate::http::{Response, Status};

use std::string::FromUtf8Error;
use thiserror::Error;
// TODO add NotFound and InternalServerError response here,
// return it in NotFound and InternalError

#[derive(Debug, Error)]
pub enum RequestError {
	#[error("request unable to be parsed as a form")]
	NotAForm,
	#[error("resource not found")]
	NotFound,
	#[error("unused method")]
	UnusedMethod,
	#[error("urldecoding error")]
	UrlDecodeErr(#[from] FromUtf8Error),
	#[error("bad request")]
	BadRequest,
	#[error("missing body in POST request")]
	MissingBody,
}

#[derive(Debug, Error)]
pub enum DatabaseError {
	#[error("failed to read from database")]
	SledError(#[from] sled::Error),
	#[error("not found in database")]
	NotInDb,
	#[error("serialization error")]
	SerError,
}

#[derive(Debug, Error)]
pub enum HandlingError {
	#[error("client error: {0}")]
	ClientError(#[from] RequestError),
	#[error("server error: {0}")]
	ServerError(#[from] DatabaseError),
	#[error("problem reading stream")]
	Io(#[from] std::io::Error),
}

impl From<DatabaseError> for Status {
	fn from(err: DatabaseError) -> Self {
		match err {
			DatabaseError::SledError(_) => Status::InternalServerError,
			DatabaseError::NotInDb => Status::BadRequest,
			DatabaseError::SerError => Status::InternalServerError,
		}
	}
}

impl From<RequestError> for Status {
	fn from(err: RequestError) -> Self {
		match err {
			RequestError::NotAForm => Status::BadRequest,
			RequestError::NotFound => Status::NotFound,
			RequestError::UnusedMethod => Status::BadRequest,
			RequestError::UrlDecodeErr(_) => Status::BadRequest,
			RequestError::BadRequest => Status::BadRequest,
			RequestError::MissingBody => Status::BadRequest,
		}
	}
}

impl From<HandlingError> for Response {
	fn from(err: HandlingError) -> Self {
		let status = match err {
			HandlingError::ClientError(e) => Status::from(e),
			HandlingError::ServerError(e) => Status::from(e),
			HandlingError::Io(_) => Status::InternalServerError,
		};
		Response::new(status, vec![], status.message().into())
	}
}
