mod errors;
mod http;
mod post;

use crate::errors::{HandlingError, RequestError};
use crate::http::{Form, Request, Response, Status};
use crate::post::{Post, Thread};

use std::io;
use std::io::{BufRead, BufReader, Read, Write};
use std::net::{TcpListener, TcpStream};
use std::path::PathBuf;

use bincode::{deserialize, serialize};
use errors::DatabaseError;
use html_escape::encode_text;
use http::Method;
use lazy_static::lazy_static;
use log::*;
use simplelog::*;
use structopt::StructOpt;
use threadpool::ThreadPool;

// use this instead of hard-coding
//type ID_TYPE = u32;
pub const INDEX: &str = include_str!("www/index.html");
pub const FAVICON: &[u8] = include_bytes!("www/favicon.ico");
pub const STYLE: &str = include_str!("www/style.css");
pub const FAQ: &str = include_str!("www/faq.html");

// represents command line arguments
#[derive(StructOpt, Debug)]
struct Opt {
	#[structopt(short, long, default_value = "8000")]
	port: u16,

	#[structopt(short, long, default_value = "data")]
	database: PathBuf,

	#[structopt(short, long, default_value = "999999999")]
	max_id: u32,
}

// get command line arguments and make them static
// safe because they're read-only
lazy_static! {
	// parse command line arguments
	static ref OPT: Opt = Opt::from_args();
	static ref DB: sled::Db = sled::open(&OPT.database).expect("failed to open db");
	static ref THREADS: sled::Tree = DB.open_tree(b"threads").expect("failed to initialize threads");
	static ref POSTS: sled::Tree = DB.open_tree(b"posts").expect("failed to intialize posts");
}

/// increments the  id stored inside the db
fn increment(old: Option<&[u8]>) -> Option<Vec<u8>> {
	let number = match old {
		Some(s) => {
			let buf: [u8; 4] = s.try_into().unwrap();
			u32::from_be_bytes(buf) + 1
			// FIXME
		}
		None => 0,
	};
	Some(number.to_be_bytes().to_vec())
}

// DATABASE ACCESS FUNCTIONS
/// returns post from id
fn get_post(id: u32) -> Result<Post, DatabaseError> {
	POSTS
		.get(id.to_be_bytes())?
		.ok_or(DatabaseError::NotInDb)
		.and_then(|x| deserialize::<Post>(&x).map_err(|_| DatabaseError::SerError))
}

/// returns thread from id
fn get_thread(id: u32) -> Result<Vec<Post>, DatabaseError> {
	THREADS
		.get(id.to_be_bytes())?
		.ok_or(DatabaseError::NotInDb)
		.and_then(|x| deserialize::<Thread>(&x).map_err(|_| DatabaseError::SerError))?
		.into_iter()
		.map(|x| get_post(x))
		.filter(|x| x.is_ok())
		.collect()
	//	let thread = THREADS.get(id.to_be_bytes()).and_then()
}

/// generates next id
fn next_id() -> Result<u32, DatabaseError> {
	DB.update_and_fetch(b"id", increment)
		.map(|x| match x {
			Some(s) => {
				let buf: [u8; 4] = (*s).try_into().unwrap();
				u32::from_be_bytes(buf)
			}
			None => 0u32,
		})
		.map_err(|e| DatabaseError::from(e))
}

/// lists all threads
fn list_threads() -> Result<Vec<Post>, DatabaseError> {
	THREADS
		.iter()
		.take_while(|x| x.is_ok())
		.map(|x| x.unwrap().0)
		.map(|x| {
			POSTS.get(x).map(|x| match x {
				Some(s) => s,
				None => {
					error!("couldn't find thread op among posts");
					panic!("unrecoverable error");
				}
			})
		})
		.map(|x| {
			x.map_err(|e| DatabaseError::from(e))
				.and_then(|i| deserialize::<Post>(&i).map_err(|_| DatabaseError::SerError))
		})
		.collect()
}

// NOTE worst out of everything, but I guess it's not so bad
fn add_post(mut post: Post, thread_id: u32) -> Result<(), DatabaseError> {
	let id = next_id()?;
	post.id = id;
	let mut thread = deserialize::<Thread>(
		&THREADS
			.get(thread_id.to_be_bytes())?
			.ok_or(DatabaseError::NotInDb)?,
	)
	.map_err(|_| DatabaseError::SerError)?;
	thread.push(post.id);
	let thread = serialize(&thread).map_err(|_| DatabaseError::SerError)?;
	THREADS.insert(thread_id.to_be_bytes(), thread)?;

	POSTS.insert(
		id.to_be_bytes(),
		serialize(&post).map_err(|_| DatabaseError::SerError)?,
	)?;
	Ok(())
}

fn delete_post(id: u32) -> Result<(), DatabaseError> {
	POSTS.remove(id.to_be_bytes())?;
	if THREADS.contains_key(id.to_be_bytes())? {
		THREADS.remove(id.to_be_bytes())?;
	} else {
	}
	Ok(())
}

fn add_thread(mut post: Post) -> Result<(), DatabaseError> {
	let id = next_id()?;
	let thread = Thread::new(id);
	post.id = id;
	THREADS.insert(
		id.to_be_bytes(),
		serialize(&thread).map_err(|_| DatabaseError::SerError)?,
	)?;
	POSTS.insert(
		id.to_be_bytes(),
		serialize(&post).map_err(|_| DatabaseError::SerError)?,
	)?;
	Ok(())
}

// END DATABASE ACCESS FUNCTIONS

/// top level handling of requests
fn handle(mut reader: BufReader<&mut TcpStream>) -> Result<Response, HandlingError> {
	let mut request: Request = reader
		.by_ref()
		.lines()
		.take_while(|l| l.is_ok())
		.map(|l| l.unwrap())
		.take_while(|l| !l.is_empty())
		.collect::<Vec<_>>()
		.try_into()?;

	if let Some(s) = request
		.headers
		.iter()
		.find(|(a, _)| a.to_lowercase() == "content-length")
	{
		if request.method == Method::Post {
			let body: Vec<u8> = reader
				.bytes()
				.take(s.1.parse::<usize>().unwrap())
				.collect::<Result<Vec<u8>, io::Error>>()
				.unwrap();
			// FIXME deunwrap
			// TODO handle body too large
			request.add_body(body);
		}
	}
	//dbg!(&request);

	match request.method {
		Method::Get => get(&request.uri),
		Method::Post => post(request),
		// TODO check admin hash
		Method::Delete => delete(&request.uri),
	}
}

/// get "/" returns head of every thread
/// or returns posts from thread with id in uri
fn get(path: &str) -> Result<Response, HandlingError> {
	match path {
		// list threads in this case
		"/" => {
			// FIXME absolute unwrap hell
			let ops = list_threads()?
				.iter()
				.fold(String::from(""), |a, b| format!("{a}\n{b}"));
			let c = content("index", &ops);

			Ok(Response::new(
				Status::Ok,
				vec![("content-type", "text/html; charset=utf-8")],
				c.into(),
			))
		}
		// TODO favicon.ico
		"/css" => Ok(Response::new(
			Status::Ok,
			vec![("content-type", "text/css; charset=utf-8")],
			String::from(STYLE).into(),
		)),
		"/faq" => Ok(Response::new(
			Status::Ok,
			vec![("content-type", "text/html; charset=utf-8")],
			String::from(FAQ).into(),
		)),
		"/favicon.ico" => Ok(Response::new(
			Status::Ok,
			vec![("content-type", "image/x-icon")],
			FAVICON.to_vec(),
		)),

		// list specific thread here
		s => {
			let id = s
				.trim_start_matches("/")
				.split("/")
				.next()
				.unwrap()
				.parse::<u32>()
				.map_err(|_| RequestError::NotFound)?;
			if s.ends_with("img") {
				Ok(Response::new(
					Status::Ok,
					vec![],
					get_post(id)?.img.ok_or(RequestError::NotFound)?,
				))
			} else {
				let c = get_thread(id)?
					.iter()
					.fold(String::from(""), |a, b| format!("{a}\n{b}"));
				let c = content(&id.to_string(), &c);
				Ok(Response::new(
					Status::Ok,
					vec![("content-type", "text/html; charset=utf-8")],
					c.into(),
				))
			}
		}
	}
}

fn post(request: Request) -> Result<Response, HandlingError> {
	// TODO check content-type

	let form = Form::try_from(&request)?;

	let mut post = Post::new(0, encode_text(form.content.trim()).replace("\r\n", "<br>"));
	post.img = form.image;

	match request.uri.as_str() {
		// means we wish to create a new thread
		"/" => {
			add_thread(post)?;
			info!("Created new thread");

			// TODO add redirect
			Ok(Response::new(
				Status::SeeOther,
				vec![("location", "/")],
				vec![],
			))
		}

		// means we wish to post in specific thread
		s => {
			let id = s
				.trim_start_matches('/')
				.split('/')
				.next()
				.unwrap()
				.parse::<u32>()
				.map_err(|_| RequestError::NotFound)?;
			add_post(post, id)?;

			info!("Added new post to thread {id}");

			// TODO add redirect
			Ok(Response::new(
				Status::SeeOther,
				vec![("location", s)],
				vec![],
			))
		}
	}
}

fn delete(path: &str) -> Result<Response, HandlingError> {
	let id = path
		.trim_start_matches("/")
		.parse::<u32>()
		.map_err(|_| RequestError::NotFound)?;
	delete_post(id)?;
	Ok(Response::new(
		Status::SeeOther,
		vec![("location", "/")],
		vec![],
	))
}

fn content(name: &str, main: &str) -> String {
	INDEX.replace("{name}", name).replace("{}", main)
}

fn main() {
	// setup logger
	TermLogger::init(
		LevelFilter::Info,
		Config::default(),
		TerminalMode::Mixed,
		ColorChoice::Auto,
	)
	.expect("failed to setup logger");

	// bind listener to local adress and port
	let listener =
		TcpListener::bind(("127.0.0.1", OPT.port)).expect("Cannot bind to specified port.");

	info!("listening on port {}", OPT.port);

	// create threadpool for incoming requests
	let pool = ThreadPool::new(num_cpus::get());

	// wait for requests
	for stream in listener.incoming() {
		let mut stream = match stream {
			Ok(s) => s,
			Err(e) => {
				warn!("Failed connection: {}", e);
				continue;
			}
		};

		pool.execute(move || {
			let reader = BufReader::with_capacity(2 << 20, &mut stream); // 2^20, should be enough for images

			let response = match handle(reader) {
				Ok(s) => s,
				Err(e) => {
					warn!("{e}");
					Response::from(e)
				}
			};

			// handle request
			stream.write_all(&response.respond()).unwrap();
		});
	}
}
